results_leukemia = dict(
    has_match=[
        True,
        False,
        True,
        True,
    ],
    detailled_status="PRESENT",
    assign=None,
    texts=[
        "Sydrome myéloprolifératif",
        "Sydrome myéloprolifératif bénin",
        "Patient atteint d'une LAM",
        "Une maladie de Vaquez",
    ],
)
