results_congestive_heart_failure = dict(
    has_match=[
        True,
        True,
        False,
        True,
        False,
    ],
    detailled_status="PRESENT",
    assign=None,
    texts=[
        "Présence d'un oedème pulmonaire",
        "Le patient est équipé d'un pace-maker",
        "Un cardiopathie non décompensée",  # no decompensation
        "Insuffisance cardiaque",
        "Insuffisance cardiaque minime",  # minimal severity
    ],
)
