results_hemiplegia = dict(
    has_match=[
        True,
        True,
        True,
    ],
    detailled_status="PRESENT",
    assign=None,
    texts=[
        "Patient hémiplégique",
        "Paralysie des membres inférieurs",
        "Patient en LIS",  # locked-in syndrom
    ],
)
