results_dementia = dict(
    has_match=[
        True,
        True,
        False,
        True,
    ],
    detailled_status="PRESENT",
    assign=None,
    texts=[
        "D'importants déficits cognitifs",
        "Patient atteint de démence",
        "On retrouve des anti-SLA",  # antibody
        "Une maladie de Charcot",
    ],
)
