results_lymphoma = dict(
    has_match=[
        True,
        True,
        True,
        False,
    ],
    detailled_status="PRESENT",
    assign=None,
    texts=[
        "Un lymphome de Hodgkin.",
        "Atteint d'un Waldenstörm",
        "Un LAGC",
        "anti LAGC: 10^4/mL",  # Dosage
    ],
)
