results_copd = dict(
    has_match=[
        True,
        True,
        True,
        False,
        False,
        True,
    ],
    detailled_status="PRESENT",
    assign=None,
    texts=[
        "Une fibrose interstitielle diffuse idiopathique",
        "Patient atteint de pneumoconiose",
        "Présence d'une HTAP.",
        "On voit une hypertension pulmonaire minime",
        "La patiente a été mis sous oxygénorequérance",  # Ponctual: not extracted
        "La patiente est sous oxygénorequérance au long cours",
    ],
)
