results_cerebrovascular_accident = dict(
    has_match=[
        False,
        True,
        True,
        False,
        True,
        True,
        True,
    ],
    detailled_status="PRESENT",
    assign=None,
    texts=[
        "Patient hospitalisé à AVC.",
        "Hospitalisation pour un AVC.",
        "Saignement intracranien",
        "Thrombose périphérique",
        "Thrombose sylvienne",
        "Infarctus cérébral",
        "Soigné via un thrombolyse",
    ],
)
