results_myocardial_infarction = dict(
    has_match=[
        True,
        False,
        True,
        False,
        True,
    ],
    detailled_status="PRESENT",
    assign=None,
    texts=[
        "Une cardiopathie ischémique",
        "Une cardiopathie non-ischémique",
        "Présence d'un stent sur la marginale",
        "Présence d'un stent périphérique",
        "infarctus du myocarde",
    ],
)
