results_connective_tissue_disease = dict(
    has_match=[
        True,
        True,
        False,
        True,
        True,
    ],
    detailled_status="PRESENT",
    assign=None,
    texts=[
        "Présence d'une sclérodermie.",
        "Patient atteint d'un lupus.",
        "Présence d'anticoagulants lupiques,",
        "Il y a une MICI.",
        "Syndrome de Raynaud",
    ],
)
