results_peptic_ulcer_disease = dict(
    has_match=[
        True,
        True,
        False,
        True,
    ],
    detailled_status="PRESENT",
    assign=None,
    texts=[
        "Beaucoup d'ulcères gastriques",
        "Présence d'UGD",
        "La patient à des ulcères",
        "Au niveau gastrique: " + 5 * "blabla " + "quelques ulcères",
    ],
)
