from typing import List

termination: List[str] = [
    "et",
    "bien que",
    "même si",
    "mais",
    "or",
    "alors que",
    "sauf",
    "cependant",
    "pourtant",
    "cause de",
    "source de",
    "hormis",
    "car",
    "parce que",
    "pourtant",
    "puisque",
    "ni",
    "en raison de",
    "qui",
    "que",
    "ainsi que",
    "avec",
    "toutefois",
    "en dehors",
    "dans le cadre",
    "du fait",
    ".",
    ",",
    ";",
    "...",
    "…",
    "(",
    ")",
    '"',
]
