results_solid_tumor = dict(
    has_match=[
        True,
        True,
        False,
        True,
        True,
        True,
        True,
    ],
    detailled_status=[
        "LOCALIZED",
        "LOCALIZED",
        None,
        "METASTASIS",
        "METASTASIS",
        "LOCALIZED",
        "METASTASIS",
    ],
    assign=None,
    texts=[
        "Présence d'un carcinome intra-hépatique.",
        "Patient avec un K sein.",
        "Il y a une tumeur bénigne",
        "Tumeur métastasée",
        "Cancer du poumon au stade 4",
        "Cancer du poumon au stade 2",
        "Présence de nombreuses lésions secondaires",
    ],
)
