results_liver_disease = dict(
    has_match=4 * [True],
    detailled_status=[
        "MILD",
        "MILD",
        "MODERATE_TO_SEVERE",
        "MODERATE_TO_SEVERE",
    ],
    assign=None,
    texts=[
        "Il y a une fibrose hépatique",
        "Une hépatite B chronique",
        "Le patient consulte pour une cirrhose",
        "Greffe hépatique.",
    ],
)
