from typing import List

pseudo: List[str] = [
    "aucun changement",
    "aucun doute",
    "aucune hésitation",
    "aucune diminution",
    "permet pas d'",
    "permet pas de",
    "permet pas d'écarter",
    "permet pas d'éliminer",
    "permet pas d'exclure",
    "permet pas d'infirmer",
    "permet pas d'interdire",
    "permet pas de nier",
    "permet pas de réfuter",
    "permet pas de rejeter",
    "ne permet pas d'",
    "ne permet pas de",
    "ne permet pas d'écarter",
    "ne permet pas d'éliminer",
    "ne permet pas d'exclure",
    "ne permet pas d'infirmer",
    "ne permet pas d'interdire",
    "ne permet pas de nier",
    "ne permet pas de réfuter",
    "ne permet pas de rejeter",
    "n'exclut pas",
    "non négligeable",
    "pas d'amélioration",
    "pas d'augmentation",
    "pas d'autre",
    "pas de changement",
    "pas de diminution",
    "pas de doute",
    "pas écarté",
    "pas écartée",
    "pas écartées",
    "pas exclu",
    "pas exclue",
    "pas exclues",
    "pas exclus",
    "pas immunisé",
    "pas immunisée",
    "pas immunisés",
    "pas immunisées",
    "sans amélioration",
    "sans aucun doute",
    "sans augmentation",
    "sans certitude",
    "sans changement",
    "sans diminution",
    "sans doute",
    "sans être certain",
]

preceding: List[str] = [
    "à la place de",
    "absence",
    "aucun",
    "aucune",
    "aucunes",
    "aucuns",
    "décline",
    "décliné",
    "dépourvu",
    "dépourvue",
    "dépourvues",
    "dépourvus",
    "disparition de",
    "disparition des",
    "élimination",
    "exclusion",
    "excluent",
    "exclut",
    "impossibilité de",
    "immunisé",
    "immunisée",
    "immunisés",
    "immunisées",
    "incompatible avec",
    "incompatibles avec",
    "infirmation de",
    "infirmation du",
    "jamais",
    "négatif pour",
    "ni",
    "niant",
    "nie",
    "nié",
    "non",
    "nullement",
    "pas",
    "permet d'exclure",
    "peu d'argument en",
    "peu d'argument pour",
    "peu d'arguments en",
    "peu d'arguments pour",
    "plus d'aspect de",
    "sans",
    "symptôme atypique",
]

preceding_regex = [
    # ne (up to 3 words separated by spaces or newlines) pas/point/...
    r"\bne\b(?=[ \n]*(?:\w*[ \n]*){3}(?:pas|point|ni|aucun|jamais|rien))"
]

following: List[str] = [
    ":0",
    ": 0",
    ":non",
    ": non",
    "absent",
    "absente",
    "absentes",
    "absents",
    "dépourvu",
    "dépourvue",
    "dépourvues",
    "dépourvus",
    "disparaissent",
    "disparait",
    "élimination",
    "est exclu",
    "est exclue",
    "immunisé",
    "immunisée",
    "immunisés",
    "immunisées",
    "impossible",
    "improbable",
    "négatif",
    "négatifs",
    "négative",
    "négatives",
    "négligeable",
    "négligeables",
    "nié",
    "niée",
    "non",
    "pas nécessaire",
    "peu probable",
    "sont exclues",
    "sont exclus",
]

verbs: List[str] = [
    "écarter",
    "éliminer",
    "exclure",
    "infirmer",
    "interdire",
    "nier",
    "réfuter",
    "rejeter",
]
