HEART = [
    r"cardi",
    r"coronari",
    r"coronair",
    r"\bcd\b",
    r"\biva\d?\b",
    r"\bivp\d?\b",
    r"\bivg\d?\b",
    r"\bivd\d?\b",
    r"intra.?va\b",
    r"intra.?vp\b",
    r"intra.?vg\b",
    r"intra.?vd\b",
    r"circonflexe",
    r"\bcx\b",
    r"marginale",
    r"\bmg\b",
    r"\bdiago",
    r"\brvp\b",
    r"myocard",
    "apical",
    "septal",
    "ventricul",
    "coeur",
    "cœur",
    "auriculaire",
    "parietal",
    "septum",
]

BRAIN = [
    r"cerveau",
    r"cereb",
    r"cran",
    r"v4",
    r"m1",
    r"aica",
    r"\bpica",
    r"basilaire",
    r"polygone de willis",
    r"cercle de willis",
    r"sylvien",
    r"arachnoi",
    r"meninge",
    r"dura(?:l|ux)",
    r"puncti",
    r"front",
    r"tempo",
    r"occipi",
    r"parieta",
    # r"segment", Too generic
]

PERIPHERAL = [
    "pied",
    "main",
    r"\bmi\b",
    r"\bmig\b",
    r"\bmid\b",
    "membre",
    "jambe",
    "bras",
    "doigt",
    "digital",
    "orteil",
]

ASYMPTOMATIC = [
    r"asympto",
    r"sans.decompens",
    r"non.decompens",
]
