results_peripheral_vascular_disease = dict(
    has_match=[
        True,
        True,
        False,
        True,
        False,
        False,
        True,
        False,
        True,
        True,
        False,
        True,
        False,
    ],
    detailled_status="PRESENT",
    assign=None,
    texts=[
        "Un AOMI",
        "Présence d'un infarctus rénal",
        "Une angiopathie cérébrale",
        "Une angiopathie",
        "Une thrombose cérébrale",
        "Une thrombose des veines superficielles",
        "Une thrombose",
        "Effectuer un bilan pre-trombose",
        "Une ischémie des MI est remarquée.",
        "Plusieurs cas d'EP",
        "Effectuer des cures d'EP",
        "Le patient est hypertendu",  # Echange plasmatique
        "Une hypertension portale",
    ],
)
