results_aids = dict(
    has_match=[
        True,
        False,
        True,
        True,
    ],
    detailled_status=[
        "PRESENT",
        None,
        "PRESENT",
        "PRESENT",
    ],
    assign=None,
    texts=[
        "Patient atteint du VIH au stade SIDA.",
        "Patient atteint du VIH.",
        "Il y a un VIH avec coinfection pneumocystose",
        "Présence d'un VIH stade C",
    ],
)
